@@ -286,13 +286,23 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
         Layer 2+: Between chunk interconnections (skip connections possible)
     """
     segids = toiter(segids)
-    if isinstance(segids, np.ndarray):
-      segids = segids.tolist()
+    input_segids = np.array(segids, dtype=self.meta.dtype)
 
-    try:
-      segids.remove(0) # background segid
-    except ValueError:
-      pass
+    if input_segids.size == 0:
+      return np.array([], dtype=self.meta.dtype)
+
+    segids = fastremap.unique(input_segids)
+
+    base_remap = { 0: 0 }
+    # skip ids that are already root IDs
+    for segid in segids:
+      if self.meta.decode_layer_id(segid) == self.meta.n_layers:
+        base_remap[segid] = segid
+
+    segids = np.array(
+      [ segid for segid in segids if segid not in base_remap ], 
+      dtype=self.meta.dtype
+    )
 
     timestamp = to_unix_time(timestamp)
 
@@ -313,7 +323,10 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
         + ", ".join([ str(_) for _ in self.meta.supported_api_versions ])
       )
 
-    return np.array(roots, dtype=self.meta.dtype)
+    for segid, root_id in zip(segids, roots):
+      base_remap[segid] = root_id
+
+    return fastremap.remap(input_segids, base_remap)
 
   def get_chunk_mappings(self, chunk_id, timestamp=None):
     """
@@ -369,6 +382,9 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
     return chunk_mappings
 
   def _get_roots_v1(self, segids, timestamp, binary=False, stop_layer=None):
+    if len(segids) == 0:
+      return []
+
     args = {}
 
     headers = {}
@@ -412,6 +428,9 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
       return orjson.loads(response.content)['root_ids']
 
   def _get_roots_legacy(self, segids, timestamp):
+    if len(segids) == 0:
+      return []
+
     args = {}
     if timestamp is not None:
       args['timestamp'] = timestamp
