@@ -152,3 +152,48 @@ def test_get_mesh_order_stability():
     assert len(first_mesh.vertices) == len(next_mesh.vertices)
     assert np.all(first_mesh.vertices == next_mesh.vertices)
     assert np.all(first_mesh.faces == next_mesh.faces)
+
+@pytest.mark.parametrize("vqb", (10,16))
+def test_stored_model_quantization(vqb):
+  from cloudvolume.datasource.precomputed.mesh import multilod
+
+  chunk_shape = [200, 200, 200]
+  grid_origin = np.random.randint(0, 101, size=(3,)).astype(np.float32)
+  vertices = grid_origin + np.random.uniform(0, 199, size=(1000,3))
+
+  manifest = multilod.MultiLevelPrecomputedMeshManifest(
+    segment_id=18, 
+    chunk_shape=chunk_shape, 
+    grid_origin=grid_origin, 
+    num_lods=1, 
+    lod_scales=[ 1 ], 
+    vertex_offsets=[[0,0,0]],
+    num_fragments_per_lod=[1], 
+    fragment_positions=[[[0,0,0]]], 
+    fragment_offsets=[0],
+  )
+
+  lod = 0
+  kwargs = {
+    "lod": lod,
+    "vertex_quantization_bits": vqb, 
+    "frag": 0
+  }
+
+  quantized_verts = multilod.to_stored_model_space( 
+    vertices, manifest, **kwargs
+  )
+  restored_verts1 = multilod.from_stored_model_space(
+    quantized_verts, manifest, **kwargs
+  )
+  quantized_verts = multilod.to_stored_model_space( 
+    restored_verts1, manifest, **kwargs
+  )
+  restored_verts2 = multilod.from_stored_model_space(
+    quantized_verts, manifest, **kwargs
+  )
+
+  precision = float(np.max(manifest.chunk_shape) / (2**vqb))
+  max_error = float(np.max(np.abs(restored_verts1 - vertices)))
+  assert max_error < precision
+  assert np.all(np.isclose(restored_verts1, restored_verts2))
\ No newline at end of file
