@@ -1,6 +1,8 @@
 import pytest 
 import sys
 
+import numpy as np
+
 import cloudvolume
 
 # Basic test of sharded meshes
@@ -29,6 +31,13 @@ def test_get_sharded_mesh(hemibrain_vol):
   meshes = hemibrain_vol.mesh.get([511271574, 360284300], lod=3)
   assert len(meshes[511271574].faces) == 50501
 
+  mesh = meshes[511271574]
+  
+  hemibrain_vol.mesh.transform[:,3] = 1
+
+  double_mesh = hemibrain_vol.mesh.get(511271574, lod=3)[511271574]
+  assert (mesh.vertices + 1 == double_mesh.vertices).all()
+
 @pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3")
 def test_get_sharded_mesh_invalid_lod(hemibrain_vol):
   try:
